import makeEventHandler from "./events.js";
// import { materials as defaultMaterials, blendMaterials } from './materials.js';

function makeRenderer(canvas = {}) {
  const events = makeEventHandler();

  let viewBox = [];
  let renderedObjects = [];
  let transformChanged = true;
  let batchedPath = new Path2D();
  let renderedObjectsChange = true;

  const ctx = canvas.getContext("2d");
  if (!ctx) {
    throw "Impossible to get canvas 2d context.";
  }
  let transform = ctx.getTransform();

  const pickingCtx = setOffscreenContext(canvas);
  pickingCtx.globalAlpha = 1;
  let pickingImageData = pickingCtx.getImageData(0, 0, pickingCtx.canvas.width, pickingCtx.canvas.height);

  const offscreenCtx = setOffscreenContext(canvas);
  let displayImageData = offscreenCtx.getImageData(0, 0, offscreenCtx.canvas.width, offscreenCtx.canvas.height);

  const drawArea = {
    get width() {
      return canvas.width;
    },
    get height() {
      return canvas.height;
    }
  };

  resize();

  function resize() {
    const pickingCanvas = pickingCtx.canvas;
    const offscreenCanvas = offscreenCtx.canvas;
    const drawAreaWidth = canvas.clientWidth * window.devicePixelRatio;
    const drawAreaHeight = canvas.clientHeight * window.devicePixelRatio;
    if (canvas.width !== drawAreaWidth || canvas.height !== drawAreaHeight) {
      const dx = drawAreaWidth - canvas.width;
      const dy = drawAreaHeight - canvas.height;
      canvas.width = pickingCanvas.width = offscreenCanvas.width = drawAreaWidth;
      canvas.height = pickingCanvas.height = offscreenCanvas.height = drawAreaHeight;
      events.emit("resize", { dx, dy });
    }
  }

  function setObjects(objects) {
    if (!Array.isArray(objects)) {
      throw `"objects" must be an array, get ${typeof objects}.`;
    }
    const oldObjectUids = new Set(renderedObjects.map(o => o.uid));
    const newObjectUids = new Set(objects.map(o => o.uid));
    if (!areSetsEqual(oldObjectUids, newObjectUids)) {
      renderedObjects = objects;
      [offscreenCtx, pickingCtx].forEach(clear);
      drawObjectsForPicking();
      pickingImageData = pickingCtx.getImageData(0, 0, pickingCtx.canvas.width, pickingCtx.canvas.height);
      makeBatchedPath();
      drawBatchedPath();
      displayImageData = offscreenCtx.getImageData(0, 0, offscreenCtx.canvas.width, offscreenCtx.canvas.height);
    } else {
      // TODO this should be optimized and react only on state objects update
      renderedObjects = objects;
      renderedObjectsChange = true;
    }
  }

  function drawObjectsForPicking() {
    for (let i = 0; i < renderedObjects.length; i++) {
      const object = renderedObjects[i];
      if (object.pickingColor && doesCollide(object.bbox, viewBox)) {
        pickingCtx.fillStyle = object.pickingColor;
        pickingCtx.fill(object.path);
      }
    }
  }

  function makeBatchedPath() {
    batchedPath = new Path2D();
    for (let i = 0; i < renderedObjects.length; i++) {
      const object = renderedObjects[i];
      if (doesCollide(object.bbox, viewBox)) {
        batchedPath.addPath(object.path)
      }
    }
  }

  function drawBatchedPath() {
    offscreenCtx.lineWidth = 0.01; // TODO remove hardcoded material
    offscreenCtx.strokeStyle = "rgba(0, 0, 0, 1)";
    offscreenCtx.stroke(batchedPath);
  }

  function drawStyledPaths() {
    const styledObjects = renderedObjects.filter(el => el.strokeStyle.color != "rgba(0,0,0,1)"); // TODO remove hardcoded material
    for (let i = 0; i < styledObjects.length; i++) {
      const styledObject = styledObjects[i];
      offscreenCtx.lineWidth = styledObject.strokeStyle.lineWidth;
      offscreenCtx.strokeStyle = styledObject.strokeStyle.color;
      offscreenCtx.fillStyle = styledObject.fillStyle.color;
      offscreenCtx.stroke(styledObject.path);
      offscreenCtx.fill(styledObject.path);
    }
  }

  return {
    render() {
      resize();
      if (transformChanged || renderedObjectsChange) {
        console.time('render');
        [ctx, offscreenCtx, pickingCtx].forEach(clear)
        if (transformChanged) {
          [offscreenCtx, pickingCtx].forEach(context => context.setTransform(transform));
          drawObjectsForPicking();
          pickingImageData = pickingCtx.getImageData(0, 0, pickingCtx.canvas.width, pickingCtx.canvas.height);
          drawBatchedPath();
          displayImageData = offscreenCtx.getImageData(0, 0, offscreenCtx.canvas.width, offscreenCtx.canvas.height);
        } else {
          pickingCtx.putImageData(pickingImageData, 0, 0);
          offscreenCtx.putImageData(displayImageData, 0, 0);
        }
        drawStyledPaths();
        ctx.drawImage(offscreenCtx.canvas, 0, 0);
        transformChanged = false;
        renderedObjectsChange = false;
        console.timeEnd('render');
      }
    },
    setTransform(t) {
      if (transform !== t) {
        transform = t;
        const dx = transform.m41 / transform.m11;
        const dy = transform.m42 / transform.m11;
        const worldWidth = canvas.width / transform.m11;
        const worldHeight = canvas.height / transform.m11;
        viewBox = [-dx, -dy, worldWidth - dx, worldHeight - dy];
        transformChanged = true;
      }
    },
    getColorAt(position = {}) {
      const { data: color } = pickingCtx.getImageData(
        position.x,
        position.y,
        1,
        1
      );
      return color;
    },
    getDrawArea() {
      return drawArea;
    },
    setObjects,
    ...events
  };
}

function setOffscreenContext(canvas) {
  const offscreenCanvas = document.createElement("canvas");
  offscreenCanvas.width = canvas.width;
  offscreenCanvas.height = canvas.height;
  return offscreenCanvas.getContext("2d");
}

function clear(context) {
  context.save();
  context.resetTransform();
  context.clearRect(0, 0, context.canvas.width, context.canvas.height);
  context.restore();
  return context;
}

function doesCollide(a, b) {
  return !(b[0] > a[2] || b[2] < a[0] || b[3] < a[1] || b[1] > a[3]);
}

function areSetsEqual(a, b) {
  return a.size === b.size && [...a].every(x => b.has(x));
}

export default makeRenderer;
