import { makeIdGenerator, capitalCaseFirstLetter } from "./utils.js";
import makeEventHandler from "./events.js";

export default () => {
  const events = makeEventHandler();
  const models = new Map();
  const objects = new Map();
  const modelIdGenerator = makeIdGenerator();
  const objectIdGenerator = makeIdGenerator();

  function loadModel(model) {
    model.uid = modelIdGenerator();
    models.set(model.uid, model);
    const addedObjects = [];
    model.objects.forEach(object => {
      object.modelUID = model.uid;
      object.uid = objectIdGenerator();
      objects.set(object.uid, object);
      addedObjects.push(object);
    });
    events.emit("update", { objects: [...objects.values()], addedObjects });
    events.emit("model-loaded", { model });
  }

  function unloadModel(uid) {
    const model = models.get(uid);
    if (model) {
      model.objects.forEach(object => objects.delete(object.uid));
      events.emit("update", { objects: [...objects.values()], deletedObjects: model.objects });
      models.delete(uid);
      events.emit("model-unloaded", { model });
      return true;
    } else {
      return false;
    }
  }

  function getObject(uid) {
    const object = [...objects.values()].find(object => object.uid === uid);
    return object ? Object.freeze({ ...object }) : undefined;
  }

  const makeObjectsPropertySetter = property => (uids = [], value) => {
    if (!Array.isArray(uids)) throw "UIDs must be an Array";
    const updatedObjects = [];
    uids
      .map(uid => objects.get(uid))
      .filter(Boolean)
      .forEach(o => {
        o[property] = value;
        updatedObjects.push(o);
      });
    events.emit("update", {
      objects: [...objects.values()],
      updatedObjects
    });
  };

  function makeObjectsPropertySetters(properties = []) {
    return properties.reduce((o, property) => {
      o[
        `setObjects${capitalCaseFirstLetter(property)}`
      ] = makeObjectsPropertySetter(property);
      return o;
    }, {});
  }

  return {
    loadModel,
    unloadModel,
    getObject,
    getAllObjects: () => [...objects.values()],
    ...makeObjectsPropertySetters([
      "visible",
      "selected",
      "xrayed",
      "highlighted",
      "pickable"
    ]),
    ...events
  };
};
