import { materials as defaultMaterials, blendMaterials } from './materials.js';

// after this value, the color restart to rgb(0,0,0)
const maxColorId = 838860799;
// a step should be added to prevent antialiasing to leads to wrong id detection
const colorStep = 50;

function getRGBStringColorFromId(id = 0) {
  if (id < 0 || id * colorStep > maxColorId) {
    throw `Impossible to generate unique color from id ${id}.`;
  }
  const rgb = [];
  rgb.push((id * colorStep) & 0xff); // R.
  rgb.push(((id * colorStep) & 0xff00) >> 8); // G.
  rgb.push(((id * colorStep) & 0xff0000) >> 16); // B.
  return `rgb(${rgb.join(",")})`;
}

const getIdFromRGBColor = function (color) {
  if (!color || color.length < 3) {
    throw "color must be an array with at least 3 values.";
  }
  const R = color[0];
  const G = color[1];
  const B = color[2];

  return R / colorStep + (G * 256) / colorStep + (B * 256 * 256) / colorStep;
};

function applyStyle(object = {}, materials = {}) {
  const appliedMaterials = [];
  if (object.highlighted && materials.highlighted) {
    appliedMaterials.push(materials.highlighted);
  }
  if (object.selected && materials.selected) {
    appliedMaterials.push(materials.selected);
  }
  if (object.xrayed && materials.xrayed) {
    appliedMaterials.push(materials.xrayed);
  }

  if (appliedMaterials.length === 0) {
    appliedMaterials.push(materials.default);
  }
  return parseMaterial(blendMaterials(appliedMaterials));
}

function parseMaterial(material = {}) {
  const parsedMaterial = {};
  if (material.strokeStyle) {
    parsedMaterial.strokeStyle = {
      lineWidth: material.strokeStyle.lineWidth,
      color: fromUint8ArrayColorToStringColor(material.strokeStyle.color, material.strokeStyle.opacity)
    }
  }
  if (material.fillStyle) {
    parsedMaterial.fillStyle = {
      color: fromUint8ArrayColorToStringColor(material.fillStyle.color, material.fillStyle.opacity)
    }
  }
  return parsedMaterial;
}

function fromUint8ArrayColorToStringColor(color, opacity = 1) {
  return `rgba(${color[0]},${color[1]},${color[2]},${opacity})`;
}

function makeStateRenderAdapter(state, renderer) {

  const materials = { ...defaultMaterials };
  let stateObjects = [];

  function sendObjectsToRenderer() {
    const objectsToSend = stateObjects.map(toRenderObject).filter(Boolean);
    renderer.setObjects(objectsToSend.sort((a, b) => a.area > b.area ? -1 : (a.area < b.area ? 1 : 0)));
  }

  state.on("update", ({ objects }) => {
    stateObjects = objects;
    sendObjectsToRenderer();
  });

  function toRenderObject(object) {
    let renderedObject = null;
    if (object.visible) {
      renderedObject = {
        uid: object.uid,
        area: object.area,
        ...applyStyle(object, materials),
        path: object.path,
        bbox: object.bbox,
        pickingColor: object.pickable ? getRGBStringColorFromId(object.uid) : null
      };
    }
    return renderedObject;
  }

  return {
    getMaterials() {
      return materials;
    },
    getObjectAtPosition(position) {
      const color = renderer.getColorAt(position);
      const uid = getIdFromRGBColor(color);
      const object = state.getObject(uid);
      return object;
    }
  };
}

export default makeStateRenderAdapter;
