import Transform2D from "./transform2D.js";
import UI from "./ui.js";
import loadSvgModel from "./svgModelLoader.js";
import makeRenderer from "./render.js";
import makeState from "./state.js";
import makeStateRendererAdapter from "./stateRendererAdapter.js";
import makeClock from "./clock.js";
import { getQuadFromBBox } from "./utils.js";
import makeEventHandler from "./events.js";

const Viewer2D = (cfg = {}) => {
  const {
    canvas,
    rotateKey = "shift",
    scaleSpeed = 0.002,
    rotateSpeed = 0.1,
    maxScale = 1000,
    minScale = 1,
    fitViewRatio = 1,
    Matrix = DOMMatrixReadOnly
  } = cfg;
  if (!(canvas instanceof HTMLCanvasElement)) {
    throw "Canvas must be a HTMLCanvasElement";
  }
  if (typeof rotateKey !== "string") {
    throw `"rotateKey" argument must be a string, get "${typeof rotateKey}"`;
  }
  [scaleSpeed, rotateSpeed, maxScale, minScale, fitViewRatio].forEach(arg => {
    if (typeof arg !== "number") {
      throw `"${arg}" argument must be a number, get "${typeof arg}"`;
    }
  });

  const state = makeState();
  const renderer = makeRenderer(canvas);
  const stateRendererAdapter = makeStateRendererAdapter(state, renderer);

  const events = makeEventHandler();

  const transform2D = Transform2D({
    drawArea: renderer.getDrawArea(),
    maxScale,
    minScale,
    Matrix
  });

  // keep draw area center aligned with canvas center on resize
  renderer.on("resize", ({ dx, dy }) => transform2D.translate(dx / 2, dy / 2));

  const ui = UI(canvas);
  ui.on("drag", ({ dx, dy }) => transform2D.translate(dx, dy));
  ui.on("wheel", ({ dy, position, keys }) => {
    if (keys[rotateKey]) {
      const angle = dy * rotateSpeed;
      transform2D.rotate(angle, position);
    } else {
      const factor = Math.max(1 + dy * scaleSpeed, 0);
      transform2D.scale(factor, position);
    }
  });
  ui.on("click", ({ position }) => {
    const object = stateRendererAdapter.getObjectAtPosition(position);
    if (object) {
      events.emit("pick", { object, position });
    } else {
      events.emit("pick-nothing", { position });
    }
  });
  ui.on("right-click", ({ position }) => {
    const object = stateRendererAdapter.getObjectAtPosition(position);
    if (object) {
      events.emit("pick", { object, position, rightClick: true });
    } else {
      events.emit("pick-nothing", { position, rightClick: true });
    }
  });
  let lastHighlightedObject = null;
  ui.on("move", ({ position }) => {
    const object = stateRendererAdapter.getObjectAtPosition(position);
    if (object) {
      if (lastHighlightedObject) {
        if (object.uid === lastHighlightedObject.uid) {
          events.emit("hover-surface", { object, position });
        } else {
          events.emit("hover", { object, position });
          events.emit("hover-out", { object: lastHighlightedObject, position });
          lastHighlightedObject = object;
        }
      } else {
        events.emit("hover", { object, position });
        lastHighlightedObject = object;
      }
    } else {
      if (lastHighlightedObject) {
        events.emit("hover-out", { object: lastHighlightedObject, position });
        lastHighlightedObject = null;
      }
      events.emit("hover-off", { position });
    }
  });

  const clock = makeClock();
  clock.on("tick", dt => {
    renderer.setTransform(transform2D.getTransform());
    renderer.render();
    events.emit("tick", dt);
  });
  clock.start();

  async function loadSvg(url) {
    return new Promise((res, err) => {
      loadSvgModel(url).then(
        model => {
          state.loadModel(model);
          res(model);
        }
      ).catch(e => err(e));
    })
  }

  return {
    getMaterials: stateRendererAdapter.getMaterials,
    loadSvg,
    unloadModel: state.unloadModel,
    setObjectsVisible: state.setObjectsVisible,
    setObjectsSelected: state.setObjectsSelected,
    setObjectsHighlighted: state.setObjectsHighlighted,
    setObjectsXrayed: state.setObjectsXrayed,
    setObjectsPickable: state.setObjectsPickable,
    getAllObjects: state.getAllObjects,
    fitView(bbox) {
      transform2D.fitView(getQuadFromBBox(bbox));
    },
    ...events
  };
};

export default Viewer2D;
